# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
import re
import warnings

import numpy as np
import pandas as pd
import polars as pl
from marginaleffects import avg_comparisons, datagrid
from pandas.api.types import is_integer_dtype
from statsmodels.regression.linear_model import RegressionResults
from statsmodels.stats.weightstats import DescrStatsW

from mozanalysis.types import (
    ComparativeOption,
    CompareBranchesOutput,
    Estimates,
    EstimatesByBranch,
    Uplift,
)

from .classes import MozOLS

logger = logging.getLogger(__name__)


## Exceptions


class LinearModelException(Exception):
    pass


class CovariateNotFound(LinearModelException):
    pass


class UnableToAnalyze(LinearModelException):
    pass


class FailedToFitModel(LinearModelException):
    pass


class MissingModelParameter(LinearModelException):
    pass


## Functions


def _stringify_alpha(alpha: float) -> tuple[str, str]:
    """Converts a floating point alpha-level to the string
    labels of the endpoint of a confidence interval.
    E.g., 0.05 -> '0.025', '0.975'"""
    if alpha < 0.002 or alpha >= 1:
        raise ValueError("alpha must be in (0.002,1)")
    return f"{alpha/2:0.3f}", f"{1-alpha/2:0.3f}"


def _make_univariate_output(alphas: list[float]) -> Estimates:
    str_quantiles = ["0.5"]
    for alpha in alphas:
        str_quantiles.extend(_stringify_alpha(alpha))
    res = pd.Series(index=sorted(str_quantiles) + ["mean"], dtype="float")
    return res


def summarize_one_branch(branch_data: pd.Series, alphas: list[float]) -> Estimates:
    """Inferences (point estimate and confidence intervals) for
    the mean of a single branch's data. Constructs confidence
    intervals from central limit theory (uses the t-distribution)

    Parameters:
    - branch_data (pd.Series): the vector of observations from a
    single branch.
    - alphas (list[float]): the desired confidence levels

    Returns:
    - result (pd.Series): the resulting inferences on the mean. Has
    the following elements in its index (assuming `alphas=[0.01,0.05]`):
      - 'mean': the point estimate of the mean
      - '0.5': also the point estimate, included for backwards compatibility
    with prior bootstrap implementations
      - '0.005': the lower bound of the 99% confidence interval
      - '0.025': the lower bound of the 95% confidence interval
      - '0.975': the upper bound of the 95% confidence interval
      - '0.995': the upper bound of the 99% confidence interval
    """
    res = _make_univariate_output(alphas)
    dsw = DescrStatsW(branch_data)
    mean = dsw.mean
    res["0.5"] = mean  # backwards compatibility
    res["mean"] = mean
    for alpha in alphas:
        low, high = dsw.tconfint_mean(alpha)
        low_str, high_str = _stringify_alpha(alpha)
        res[low_str] = low
        res[high_str] = high
    return res


def _infer_branch_list(branches: pd.Series, branch_list: list[str] | None) -> list[str]:
    """Determine the list of distinct branches. Used so that `summarize_univariate`
    and `summarize_joint` can take an optional `branch_list` parameter."""
    if branch_list:
        return branch_list

    return list(branches.unique())


def summarize_univariate(
    data: pd.Series,
    branches: pd.Series,
    alphas: list[float],
    branch_list: list[str] | None = None,
) -> EstimatesByBranch:
    """Univariate inferences (point estimates and confidence intervals) for the
    mean of each branch's data.

    Parameters:
    - data (pd.Series): the vector of observations
    - branches (pd.Series): the vector of branch names. Expected to be the same length
    as `data`
    - branch_list (list[str]): the list of branches to perform inferences on
    - alphas (list[float]): the desired confidence levels

    Returns:
    - out (dict[str, pd.Series]): a dictionary keyed by branch label with a pandas
    series as value. The format of the series is described above in
    `summarize_one_branch`
    """

    branch_list = _infer_branch_list(branches, branch_list)

    return {
        b: summarize_one_branch(data.loc[(branches == b).values], alphas)  # type: ignore
        for b in branch_list
    }


def make_formula(target: str, ref_branch: str, covariate: str | None = None) -> str:
    """Makes a formula which defines the model to build. Includes terms for
    treatment branch and, optionally, a main effect for a (generally pre-experiment)
    covariate.

    Ex: given `target` of 'Y' and `ref_branch` of control, the formula will be:
    `Y ~ C(branch, Treatment(reference='control'))`. That is, we want to predict `Y`
    with treatment branch (`branch`) treated as a categorical predictor with
    reference level of `'control'`.

    This builds the following linear model (assuming 2 other treatment branches, t1
    and t2): `Y_i = \beta_0 + \beta_1*I(branch == 't1') + \beta_2*I(branch == 't2')`
    where `I` is the indicator function.

    Inferences on `\beta_1` are inferences of the average treatment effect (ATE) of
    branch t1. That is, the confidence interval for `\beta_1` is the confidence
    interval for the ATE of branch t1. Similarly for t2 and `\beta_2`.

    We can incorporate a single (generally pre-experiment) covariate. Adding a covariate
    of `Ypre` to the above, we'll build the following formula:
    `Y ~ C(branch, Treatment(reference='control')) + Ypre`
    which will fit the following linear model:

    `Y_i = \beta_0 + \beta_1*I(branch == 't1') + \beta_2*I(branch == 't2')
    + \beta_3* Ypre_i`

    For now, we elect to not include branch by covariate interaction terms and instead
    to perform inferences only on the population-level ATE.

    Parameters:
    - target (str): the variable of interest.
    - ref_branch (str): the name of the reference branch
    - covariate (Optional[str]): the name of a covariate to include in the model.

    Returns:
    - formula (str): the R-style formula, to be passed to statsmodels's formula API.

    """
    pattern = re.compile(r"(\(|\)|\~|\')")
    if pattern.findall(target):
        raise ValueError(f"Target variable {target} contains invalid character")
    if pattern.findall(ref_branch):
        raise ValueError(f"Reference branch {ref_branch} contains invalid character")
    if covariate is not None and pattern.findall(covariate):
        raise ValueError(f"Covariate {covariate} contains invalid character")

    formula = f"{target} ~ C(branch, Treatment(reference='{ref_branch}'))"
    if covariate is not None:
        formula += f" + {covariate}"

    return formula


def _make_joint_output(alphas: list[float], uplift: Uplift) -> Estimates:
    """Constructs an empty pandas series to hold comparative results. The series
    will be multiindexed for backwards compatability with the bootstrap results.

    Parameters:
    - alphas (list[float]): the desired confidence levels
    - uplift (Uplift): either Uplift.ABSOLUTE or Uplift.RELATIVE for inferences on the
    absolute and relative differences between branches, respectively.

    Returns:
    - series (pd.Series): the empty series. Will have keys of (uplift.value, '0.5')
    and (uplift.value, 'exp'), as well as 2 keys, one for each alpha. For more info
    on keys, see `summarize_one_branch` and `stringify_alpha` above.
    """
    str_quantiles = ["0.5", "exp"]
    for alpha in alphas:
        str_quantiles.extend(_stringify_alpha(alpha))
    str_quantiles.sort()
    index = pd.MultiIndex.from_tuples([(uplift.value, q) for q in str_quantiles])
    series = pd.Series(index=index, dtype="float")

    return series


def _extract_absolute_uplifts(
    results: RegressionResults, branch: str, ref_branch: str, alphas: list[float]
) -> Estimates:
    """Extracts inferences on absolute differences between branches from a fitted
    linear model. These are simply the point estimates and confidence intervals of the
    appropriate term in the model.

    Parameters:
    - results (RegressionResults): the fitted model.
    - branch (str): the name of the branch inferences are desired on. Must have been
    present in the training data when the model was fit.
    - ref_branch (str): the name of the reference branch.
    - alphas (list[float]): the disired confidence levels.

    Returns:
    - output (pd.Series): the set of inferences. See `_make_joint_output`.
    """
    output = _make_joint_output(alphas, Uplift.ABSOLUTE)
    parameter_name = f"C(branch, Treatment(reference='{ref_branch}'))[T.{branch}]"
    abs_uplift = Uplift.ABSOLUTE.value
    output.loc[(abs_uplift, "0.5")] = results.params[parameter_name]
    output.loc[(abs_uplift, "exp")] = results.params[parameter_name]

    for alpha in alphas:
        ci = results.conf_int(alpha=alpha)
        lower, upper = ci.loc[parameter_name]
        low_str, high_str = _stringify_alpha(alpha)
        output.loc[(abs_uplift, low_str)] = lower
        output.loc[(abs_uplift, high_str)] = upper
    return output


def _create_datagrid(
    results: RegressionResults,
    branches: list[str],
    covariate_col_label: str | None = None,
) -> pl.DataFrame:
    """
    Creates a grid of data which approximates the empirical data distribution. This is
    used to dramatically reduce the runtime/memory use of `_extract_relative_uplifts`
    when the data is large.

    Parameters:
    - results (RegressionResults): the fitted model.
    - branches (list[str]): the list of all experiment branches
    - covariate_col_label (Optional[str]): the name of the covariate used in modeling

    """
    if covariate_col_label is None:
        newdata = datagrid(model=results, grid_type="balanced", branch=branches)
    else:
        q = (
            results.model.data.frame[covariate_col_label]
            .astype(float)
            .quantile(np.arange(0, 1, 0.0001))
            .values
        )
        newdata = datagrid(
            model=results,
            grid_type="balanced",
            **{covariate_col_label: q},
            branch=branches,
        )
    return newdata


def _extract_relative_uplifts(
    results: RegressionResults,
    branch: str,
    ref_branch: str,
    alphas: list[float],
    treatment_branches: list[str],
    covariate_col_label: str | None = None,
) -> Estimates:
    """Extracts inferences on relative differences between branches from a fitted
    linear model. Unlike absolute differences, these are not simply existing parameters.
    We desire inferences on (T-C)/C = T/C-1. Since 1 is a constant, we desire inferences
    on T/C. We can accomplish this through inferences on exp(ln(T/C)) =
    exp(ln(T) - ln(C)). The ln(T)-ln(C) contrast is available in
    `marginaleffects.avg_comparisons` as `lnratioavg`.

    See [this stats.stackexchange](https://stats.stackexchange.com/a/646462/82977)
    answer for more information.

    Parameters:
    - results (RegressionResults): the fitted model.
    - branch (str): the name of the branch inferences are desired on. Must have been
    present in the training data when the model was fit.
    - ref_branch (str): the name of the reference branch.
    - alphas (list[float]): the disired confidence levels.
    - treatment_branches (list[str]): the list of treatment branches.
    - covariate_col_label (Optional[str]): the name of the covariate used in modeling

    Returns:
    - output (pd.Series): the set of inferences. See `_make_joint_output`.
    """

    output = _make_joint_output(alphas, Uplift.RELATIVE)
    branches = treatment_branches + [ref_branch]

    newdata = _create_datagrid(results, branches, covariate_col_label)

    for alpha in alphas:
        # inferences on branch/ref_branch
        ac = avg_comparisons(
            results,
            variables={"branch": [ref_branch, branch]},
            comparison="lnratioavg",
            transform=np.exp,
            conf_level=1 - alpha,
            newdata=newdata,
        )

        assert ac.shape == (
            1,
            7,
        ), "avg_comparisons result object not shaped as expected"

        low_str, high_str = _stringify_alpha(alpha)
        # subtract 1 b/c branch/reference - 1 = (branch - reference)/reference
        rel_uplift = Uplift.RELATIVE.value
        output.loc[(rel_uplift, low_str)] = ac["conf_low"][0] - 1
        output.loc[(rel_uplift, high_str)] = ac["conf_high"][0] - 1
        output.loc[(rel_uplift, "0.5")] = ac["estimate"][0] - 1
        output.loc[(rel_uplift, "exp")] = ac["estimate"][0] - 1

    return output


def fit_model(
    df: pd.DataFrame,
    target: str,
    ref_branch: str,
    treatment_branches: list[str],
    covariate: str | None = None,
    deallocate_aggressively: bool = False,
) -> RegressionResults:
    """Fits a linear regression model to `df` using the provided formula. See
    `make_formula` for a more in-depth discussion on the model structure.

    Parameters:
    - df (pd.DataFrame): the model data
    - target (str): the target column name (experimental metric of interest).
    Passed to `make_formula`.
    - ref_branch (str): the name of the reference branch. e.g. "control"
    Passed to `make_formula`.
    - treatment_branches (list[str]): the set of non-reference branch names.
    e.g. ["treatment-a", "treatment-b"].
    - covariate (Optional[str]): the column name of the pre-treatment covariate.
    Passed to `make_formula`.
    - deallocate_aggressively (bool): drop large objects as soon as possible trigger
    and garbage collect. Not normally needed in interactive use, but can help when
    called through Jetstream.

    Returns:
    - results (RegressionResults): the fitted model results object.

    Raises:
    - `FailedToFitModel` if unable to fit the model.
    - `MissingModelParameter` if one of the treatment branches, which was expected to
    have a model parameter, does not get one.
    """
    formula = make_formula(target, ref_branch, covariate)

    model = MozOLS.from_formula(formula, df)
    fell_back = False
    lin_alg_error = False
    try:
        results = model.fit()
    except np.linalg.LinAlgError:
        if covariate is None:
            # nothing we can do about this
            lin_alg_error = True

        else:
            # maybe covariate is bad in a way that we havent otherwise accounted for
            # try falling back to unadjusted inferences
            msg = "Unexpectedly fell back to unadjusted inferences"
            formula = make_formula(target, ref_branch, None)
            model = MozOLS.from_formula(formula, df)
            try:
                results = model.fit()
                logger.warning(msg)
                warnings.warn(msg, stacklevel=1)
            except np.linalg.LinAlgError:
                # well... we tried our best
                lin_alg_error = True

            fell_back = True

    if lin_alg_error:
        msg = (
            "Encountered a linear algebra error when fitting"
            f" model for target {target}"
        )
        if covariate is not None:
            msg += " even when falling back to unadjusted inferences"
        logger.warning(msg)
        warnings.warn(msg, stacklevel=1)
        raise FailedToFitModel(msg)

    if not np.isfinite(results.llf):
        msg = f"Failed to fit model for target {target}"
        if covariate is not None:
            msg += f" using covariate {covariate}"
            if fell_back:
                msg += " even after attempting to recover using unadjusted inferences"
        raise FailedToFitModel(msg)

    for branch in treatment_branches:
        param_name = f"C(branch, Treatment(reference='{ref_branch}'))[T.{branch}]"
        if param_name not in results.params:
            # this can occur if a branch does not have any non-null data
            raise MissingModelParameter(
                f"Effect for branch {branch} not found in model!"
            )

    if deallocate_aggressively:
        import gc

        # warm up the cache - necessary for confidence intervals
        results.bse  # noqa: B018
        results.remove_data()
        gc.collect()

    return results


def summarize_joint(
    df: pd.DataFrame,
    col_label: str,
    alphas: list[float],
    branch_list: list[str] | None = None,
    ref_branch_label: str = "control",
    covariate_col_label: str | None = None,
    deallocate_aggressively: bool = False,
) -> EstimatesByBranch:
    """The primary entrypoint for linear model based inferences on comparisons
    of treatment branches. Computes absolute and relative differences between
    each treatment branch relative to the reference branch (point estimates and
    confidence intervals).

    Absolute differences are the average treatment effect (ATE): T - C, for branch
    T and reference branch C.

    Relative differences are (T-C)/C for branch T and reference branch C.

    When `covariate_col_label` is passed, the covariate is included into the model and
    affects inferences on the branch differences. The current recommendation is to
    use the pre-experiment version of the metric under measurement (col_label). E.g.,
    if studying active_hours, incorporate average active_hours during the week or month
    prior to the experiment.

    Parameters:
    - df (pd.DataFrame): a cleaned set of data, ready for modeling (such as the output
    of `make_model_df`)
    - col_label (str): the target variable for which inferences are desired.
    - branch_list (list[str]): the set of all branches (treatments and controls) in the
    experiment.
    - alphas (list[str]): the desired confidence levels.
    - ref_branch_label (str): the name of the reference branch (e.g., 'control')
    - covariate_col_label (Optional[str]): the name of a covariate to include in the
    model.
    - deallocate_aggressively (bool): drop large objects as soon as possible trigger
    and garbage collect. Not normally needed in interactive use, but can help when
    called through Jetstream.

    Returns:
    - output (dict[str, pd.Series]): a dictionary keyed by (non-reference) branch,
    containing the comparative results of that branch against the reference branch.

    """

    branch_list = _infer_branch_list(df.branch, branch_list)

    treatment_branches = [b for b in branch_list if b != ref_branch_label]

    results = fit_model(
        df, col_label, ref_branch_label, treatment_branches, covariate_col_label
    )

    output = {}

    for branch in treatment_branches:
        abs_uplifts = _extract_absolute_uplifts(
            results, branch, ref_branch_label, alphas
        )

        if (
            covariate_col_label is None
            or covariate_col_label not in results.params.index
        ):
            rel_uplifts = _extract_relative_uplifts(
                results, branch, ref_branch_label, alphas, treatment_branches
            )
        else:
            rel_uplifts = _extract_relative_uplifts(
                results,
                branch,
                ref_branch_label,
                alphas,
                treatment_branches,
                covariate_col_label,
            )
        output[branch] = pd.concat([rel_uplifts, abs_uplifts])

    if deallocate_aggressively:
        import gc

        del results
        gc.collect()

    return output


def prepare_df_for_modeling(
    df: pd.DataFrame,
    target_col: str,
    threshold_quantile: float | None = None,
    covariate_col: str | None = None,
    copy: bool = True,
) -> pd.DataFrame:
    """
    Performs outlier clipping inplace and returns a view into the dataframe that can be
    used for modeling: target and covariate, if passed, are guaranteed to be non-null.

    Parameters:
    - df (pd.DataFrame): a cleaned set of data, ready for modeling (such as the output
    of `make_model_df`)
    - target_col (str): the target variable for which inferences are desired.
    - threshold_quantile (Optional[float]): the outlier threshold. See `filter_outliers`
    - covariate_col_label (Optional[str]): the name of a covariate to include in the
    model.
    - copy (bool): if True (default) returns a cleaned copy of the data. If False,
    modifies data inplace
    """
    indexer = ~df[target_col].isna()
    if covariate_col is not None:
        indexer &= ~df[covariate_col].isna()

    if copy:
        df = df.copy()

    if threshold_quantile is not None:
        threshold = df[target_col].quantile(threshold_quantile)
        if is_integer_dtype(df[target_col].dtype):
            threshold = int(np.ceil(threshold))

        df[target_col] = df[target_col].clip(upper=threshold)

    if (covariate_col is not None) and (threshold_quantile is not None):
        threshold = df[covariate_col].quantile(threshold_quantile)
        if is_integer_dtype(df[covariate_col].dtype):
            threshold = int(np.ceil(threshold))

        df[covariate_col] = df[covariate_col].clip(upper=threshold)

    return df.loc[indexer]


def _make_empty_compare_branches_output(
    ref_branch_label: str,
    branches: pd.Series,
    alphas: list[float],
    branch_list: list[str] | None = None,
) -> CompareBranchesOutput:
    """
    Constructs an empty output to be returned to Jetstream in the case of an
    expected failure mode.
    """
    out: CompareBranchesOutput = {}
    branch_list = _infer_branch_list(branches, branch_list)
    individual = {b: _make_univariate_output(alphas) for b in branch_list}
    out[ComparativeOption.INDIVIDUAL] = individual

    treatment_branches = [b for b in branch_list if b != ref_branch_label]
    comparative_empty_result = pd.concat(
        [
            _make_joint_output(alphas, Uplift.ABSOLUTE),
            _make_joint_output(alphas, Uplift.RELATIVE),
        ]
    )
    comparative = {b: comparative_empty_result.copy() for b in treatment_branches}
    out[ComparativeOption.COMPARATIVE] = comparative

    return out


def _validate_parameters(
    df: pd.DataFrame,
    col_label: str,
    ref_branch_label="control",
    covariate_col_label: str | None = None,
    threshold_quantile: float | None = None,
    alphas: list[float] | None = None,
) -> None:
    """
    Validates the analysis parameters to check for common failure cases. Intended
    to be called by `compare_branches_lm`.

    Raises `UnableToAnalyze` if `compare_branches_lm` should gracefully return an
    empty analysis output. This is the case if the failure mode is of a known type
    and we want the analyst to see the empty results, to know that analysis was
    attempted but refused. In this case, Jetstream will not throw an error that
    would trigger an alert.

    Raises `CovariateNotFound` if covariate-adjustment was requested but the
    covariate isn't present. This indicates that `compare_branches_lm` should
    fall back to unadjusted inferences.

    Raises `ValueError` if any other non-recoverable error was found. In this case,
    we desire Jetstream to log an error and alert on the issue. These are generally
    misconfigured parameters.
    """

    if col_label not in df.columns:
        # this should never happen... raise up to Jetstream
        raise ValueError(f"Target metric {col_label} not found in data")

    if np.isclose(df[col_label].std(), 0):
        # this error can be expected in the case of preenrollment analysis
        # periods for new-user experiments (e.g., onboarding) where there is
        # no preenrollment data.
        # we only need to check target, if covariate has no variation,
        # it will not be modeled
        raise UnableToAnalyze(f"Metric {col_label} has no variation!")

    if ref_branch_label not in df.branch.unique():
        raise ValueError(f"No data from reference branch {ref_branch_label} found")

    if (threshold_quantile is not None) and (
        threshold_quantile <= 0 or threshold_quantile > 1
    ):
        raise ValueError("Threshold quantile must be in (0,1]")

    if alphas:
        for alpha in alphas:
            if alpha < 0.002 or alpha >= 1:
                raise ValueError("alpha must be in (0.002,1)")

    if covariate_col_label:
        if covariate_col_label not in df.columns:
            raise CovariateNotFound(
                f"Covariate {covariate_col_label} not found in data"
            )
        if covariate_col_label == col_label:
            raise ValueError(
                f"Covariate {covariate_col_label} must be different than target {col_label}"  # noqa: E501
            )


def compare_branches_lm(
    df: pd.DataFrame,
    col_label: str,
    ref_branch_label="control",
    covariate_col_label: str | None = None,
    threshold_quantile: float | None = None,
    alphas: list[float] | None = None,
    interactive: bool = True,
    deallocate_aggressively: bool = False,
) -> CompareBranchesOutput:
    """Performs individual and comparative inferences on branches using standard
    central limit theory (t-tests and linear regressions). Can incorporate a covariate
    (`covariate_col_label`) to increase precision of comparative inferences.

    See `summarize_univariate` and `summarize_joint` for more information.

    Parameters:
    - df (pd.DataFrame): the raw, client-level data. Must have a column "branch".
    - col_label (str): the label of the response variable. Must be a column in df. Only
    non-null values are modeled.
    - ref_branch_label (str): the label of the reference branch (e.g., "control").
    - covariate_col_label (Optional[str]): the name of a covariate column to adjust for.
    If not passed, unadjusted inferences are provided.
    - threshold_quantile (Optional[float]): the outlier threshold. See `filter_outliers`
    - alphas (list[float]): the desired confidence levels. Defaults to [0.01, 0.05]
    (99% and 95% confidence) if not passed.
    - interactive (bool): When true (default), copies the modeling data to avoid
    unexpected side effects. When false (used by Jetstream), it modifies data in-place
    for reduced memory use. Passed to `prepare_df_for_modeling`
    - deallocate_aggressively (bool): drop large objects as soon as possible trigger
    and garbage collect. Not normally needed in interactive use, but can help when
    called through Jetstream.

    Returns:
    - out (dict[str, dict[str, pd.Series]]): the results. Has keys `individual` for
    univariate inferences and `comparative` for joint. See `summarize_univariate`
    and `summarize_joint` for more information on the output structure.

    """
    if alphas is None:
        alphas = [0.01, 0.05]

    try:
        _validate_parameters(
            df,
            col_label,
            ref_branch_label,
            covariate_col_label,
            threshold_quantile,
            alphas,
        )
    except CovariateNotFound:
        initial_covariate_col_label = covariate_col_label
        covariate_col_label = None
        # revalidate
        _validate_parameters(
            df,
            col_label,
            ref_branch_label,
            covariate_col_label,
            threshold_quantile,
            alphas,
        )

        logger.warn(
            f"Covariate {initial_covariate_col_label} not found, falling back to unadjusted inferences"  # noqa: E501
        )
    except UnableToAnalyze:
        # validation has found a case of degenerate data that we are not able to
        # perform analysis on. Therefore, we gracefully fail by returning an
        # empty output to Jetstream
        branch_list = _infer_branch_list(df.branch, None)
        return _make_empty_compare_branches_output(
            ref_branch_label, df.branch, alphas, branch_list
        )

    model_df = prepare_df_for_modeling(
        df, col_label, threshold_quantile, covariate_col_label, copy=interactive
    )

    branch_list = _infer_branch_list(model_df.branch, None)

    out: CompareBranchesOutput = {}
    out[ComparativeOption.INDIVIDUAL] = summarize_univariate(
        model_df[col_label], model_df.branch, alphas, branch_list=branch_list
    )
    try:
        comparative = summarize_joint(
            model_df,
            col_label,
            alphas,
            ref_branch_label=ref_branch_label,
            branch_list=branch_list,
            covariate_col_label=covariate_col_label,
            deallocate_aggressively=deallocate_aggressively,
        )
        out[ComparativeOption.COMPARATIVE] = comparative
    except FailedToFitModel as e:
        # we encountered an unrecoverable error when attempting to fit the linear model
        # as a result, we gracefully fail by returning an empty comparative result
        logger.warning(e)
        out[ComparativeOption.COMPARATIVE] = _make_empty_compare_branches_output(
            ref_branch_label, model_df.branch, alphas, branch_list
        )[ComparativeOption.COMPARATIVE]

    return out



from enum import Enum
from typing import TypeAlias

import pandas as pd


class ComparativeOption(str, Enum):
    INDIVIDUAL = "individual"
    COMPARATIVE = "comparative"


class Uplift(str, Enum):
    ABSOLUTE = "abs_uplift"
    RELATIVE = "rel_uplift"


Estimates: TypeAlias = pd.Series
BranchLabel = str
EstimatesByBranch = dict[BranchLabel, Estimates]

CompareBranchesOutput = dict[ComparativeOption, EstimatesByBranch]


class IncompatibleAnalysisUnit(ValueError):
    pass



import numpy as np
import pandas as pd
import statsmodels.formula.api as smf

# all models in this file are fit using standard statsmodels.OLS objects
# so that they can be compared against MozOLS models


class UnitTestModel:
    def __init__(self):
        self.ref_branch = "treatment-a"
        self.alphas = [0.01, 0.05]
        self.branches = (
            ["control"] * 100 + ["treatment-a"] * 100 + ["treatment-b"] * 100
        )
        self.treatment_branches = ["control", "treatment-b"]
        searches = list(range(100)) + list(range(100, 200)) + list(range(200, 300))
        self.model_df = pd.DataFrame(
            {"search_count": searches, "branch": self.branches}
        )
        self.target = "search_count"
        self.formula = "search_count ~ C(branch, Treatment(reference='treatment-a'))"
        self.results = smf.ols(self.formula, self.model_df).fit()


class UnitTestModelCovariate:
    def __init__(self):
        np.random.seed(42)
        self.ref_branch = "treatment-a"
        self.alphas = [0.01, 0.05]
        self.treatment_branches = ["control", "treatment-b"]
        branches = ["control"] * 100 + ["treatment-a"] * 100 + ["treatment-b"] * 100
        y_base = np.random.normal(loc=2, scale=0.1, size=300)
        y_pre_adj = np.random.normal(loc=0, scale=0.02, size=300)
        te = np.concatenate(
            [
                np.random.normal(loc=0, scale=0.05, size=100),
                np.random.normal(loc=0.1, scale=0.05, size=100),
                np.random.normal(loc=0.2, scale=0.05, size=100),
            ]
        )
        self.model_df = pd.DataFrame(
            {
                "search_count": y_base + te,
                "branch": branches,
                "search_count_pre": y_base + y_pre_adj,
            }
        )
        self.formula = "search_count ~ C(branch, Treatment(reference='treatment-a')) + search_count_pre"  # noqa: E501
        self.target = "search_count"
        self.covariate = "search_count_pre"
        self.results = smf.ols(self.formula, self.model_df).fit()


test_model = UnitTestModel()
test_model_covariate = UnitTestModelCovariate()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import warnings
from datetime import date
from typing import TYPE_CHECKING

import attr

from mozanalysis.bq import BigQueryContext, sanitize_table_name_for_bq
from mozanalysis.experiment import (
    AnalysisWindow,
    TimeLimits,
    TimeSeriesResult,
    add_days,
)
from mozanalysis.utils import hash_ish

if TYPE_CHECKING:
    from pandas import DataFrame

    from mozanalysis.metrics import DataSource, Metric
    from mozanalysis.segments import Segment, SegmentDataSource


@attr.s(frozen=False, slots=True)
class HistoricalTarget:
    """Query historical data.

    The methods here query data in a way compatible with the following
    principles, which are important for experiment analysis:

    Example usage (in a colab notebook)::

        from google.colab import auth
        auth.authenticate_user()
        print('Authenticated')

        from mozanalysis.config import ConfigLoader

        active_hours = ConfigLoader.get_metric("active_hours", "firefox_desktop")
        uri_count = ConfigLoader.get_metric("uri_count", "firefox_desktop")

        new_or_resurrected_v3 = ConfigLoader.get_segment("new_or_resurrected_v3",
                                                            "firefox_desktop")
        allweek_regular_v1 = ConfigLoader.get_segment("allweek_regular_v1",
                                                            "firefox_desktop")

        bq_context = BigQueryContext(
            dataset_id='mbowerman',  # e.g. mine's mbowerman
            project_id='moz-fx-data-bq-data-science'  # this is the default anyway
        )

        ht = HistoricalTarget(
            experiment_name='my_test_name',
            start_date='2021-01-01',
            num_dates_enrollment=2,
            analysis_length=4
        )

        # Run the query and get the results as a DataFrame
        res = ht.get_single_window_data(
            bq_context,
            metric_list = [
                active_hours,
                uri_count
            ],
            target_list = [
                allweek_regular_v1,
                new_or_resurrected_v3
            ]
        )

        targets_sql = ht.get_targets_sql()

        metrics_sql = ht.get_metrics_sql()

    Args:
        experiment_name (str): A simple name of this analysis, which is used
            to create table names when saving results in BigQuery.
        start_date (str): e.g. '2019-01-01'. First date for historical data to be
            retrieved.
        analysis_length (int): Number of days to include for analysis
        num_dates_enrollment (int, default 7): Number of days to consider to
            enroll clients in the analysis.
        continuous_enrollment (bool): Indicates if the analysis dates
            and enrollment should be overlap; clients that satisfy the
            target conditions at any point in the analysis will be
            included for the entire window when calculating metrics.
        app_id (str, optional): For a Glean app, the name of the BigQuery
            dataset derived from its app ID, like `org_mozilla_firefox`.

    Attributes:
        experiment_name (str): Name of the study, used in naming tables
            related to this analysis.
        start_date (str): e.g. '2019-01-01'. First date on which enrollment
            events were received.
        num_dates_enrollment (int): Number of days to consider to enroll clients
            in the analysis.
        analysis_length (int, optional): Number of days to include for analysis
    """

    experiment_name = attr.ib()
    start_date = attr.ib()
    analysis_length = attr.ib()
    num_dates_enrollment = attr.ib(default=7)
    continuous_enrollment = attr.ib(default=False)
    app_id = attr.ib(default=None)

    _metrics_sql = attr.ib(default=None)
    _targets_sql = attr.ib(default=None)

    def get_single_window_data(
        self,
        bq_context: BigQueryContext,
        metric_list: list[Metric],
        target_list: list[Segment] | None = None,
        custom_targets_query: str | None = None,
        replace_tables: bool = False,
    ) -> DataFrame:
        """Return a DataFrame containing per-client metric values.

        Also store them in a permanent table in BigQuery. The name of
        this table will be printed. Subsequent calls to this function
        will simply read the results from this table.

        Args:
            bq_context (BigQueryContext): BigQuery configuration and client.
            metric_list (list of mozanalysis.metric.Metric): The metrics
                to analyze.
            target_list (list of mozanalysis.segments.Segment): The targets
                that define clients to be included in the analysis, based on
                inclusion in a defined user segment. For each Segment included
                in the list, client IDs are identified in the
                SegmentDataSource that based on the select_for statement in
                the Segment. Client IDs that satisfy the select_for
                statement in every Segment in the target_list will be
                returned by the query.
            custom_targets_query (str): A full SQL query to be used
                in the main query; must include a client_id column
                to join to metrics tables and an enrollment_date:

                N.B. this query's results must be uniquely keyed by
                (client_id, enrollment_date), or else your results will be
                subtly wrong.
            replace_tables (bool): If True, delete tables that exist in
                BigQuery with the same name; used to rerun analyses under
                the same experiment name with different settings

        Returns:
            A pandas DataFrame of experiment data. One row per ``client_id``.
            Some metadata columns, then one column per metric in
            ``metric_list``, and one column per sanity-check metric.
            Columns (not necessarily in order):

                * client_id (str): Not necessary for "happy path" analyses.
                * Analysis window (int): Start and end of analysis window, in
                  days of entire analysis included in the window
                * [metric 1]: The client's value for the first metric in
                  ``metric_list``.
                * ...
                * [metric n]: The client's value for the nth (final)
                  metric in ``metric_list``.
        """
        if not (custom_targets_query or target_list):
            raise ValueError(
                "Either custom_target_query or target_list must be provided"
            )

        # validate metric_list and target_list are from the same source
        # filter out el.app_name is None, there are cases where metric-hub
        # was not used and it's up to users to ensure the sources match
        metric_list_sources = {el.app_name for el in metric_list if el.app_name}
        target_list_sources = {el.app_name for el in target_list if el.app_name}

        if len(metric_list_sources) > 1:
            warnings.warn("metric_list contains multiple metric-hub apps", stacklevel=1)

        if len(target_list_sources) > 1:
            warnings.warn("target_list contains multiple metric-hub apps", stacklevel=1)

        if (
            metric_list_sources
            and target_list_sources
            and metric_list_sources != target_list_sources
        ):
            warnings.warn(
                "metric_list and target_list metric-hub apps do not match",
                stacklevel=1,
            )

        last_date_full_data = add_days(
            self.start_date, self.num_dates_enrollment + self.analysis_length - 1
        )

        today = date.today().strftime("%Y-%m-%d")
        if last_date_full_data >= today:
            raise ValueError(
                f"Based on the start date, {self.start_date}"
                + f", with {self.num_dates_enrollment} days of enrollment "
                + f"and analysis of length {self.analysis_length} days, "
                + f"the last day of analysis is {last_date_full_data}"
                + ", which is in the future."
            )

        if self.continuous_enrollment:
            time_limits = ContinuousEnrollmentTimeLimits.for_single_analysis_window(
                self.start_date,
                self.analysis_length,
            )

        else:
            time_limits = TimeLimits.for_single_analysis_window(
                self.start_date,
                last_date_full_data,
                1,
                self.analysis_length,
                self.num_dates_enrollment,
            )

        self._targets_sql = self.build_targets_query(
            time_limits=time_limits,
            target_list=target_list,
            custom_targets_query=custom_targets_query,
        )

        targets_table_name = sanitize_table_name_for_bq(
            "_".join(
                [
                    time_limits.last_date_data_required,
                    "targets",
                    self.experiment_name,
                    hash_ish(self._targets_sql),
                ]
            )
        )

        bq_context.run_query(self._targets_sql, targets_table_name, replace_tables)

        self._metrics_sql = self.build_metrics_query(
            metric_list=metric_list,
            time_limits=time_limits,
            targets_table=bq_context.fully_qualify_table_name(targets_table_name),
        )

        full_res_table_name = sanitize_table_name_for_bq(
            "_".join(
                [
                    time_limits.last_date_data_required,
                    self.experiment_name,
                    hash_ish(self._metrics_sql),
                ]
            )
        )

        output = bq_context.run_query(
            self._metrics_sql, full_res_table_name, replace_tables
        ).to_dataframe()

        for metric_obj in metric_list:
            if all(output[metric_obj.name] == 0):
                warnings.warn(
                    (
                        f"Metric {metric_obj.name} is all 0, which may indicate"
                        + " segments and metric do not have a common app"
                    ),
                    stacklevel=1,
                )
        return output

    def get_time_series_data(
        self,
        bq_context: BigQueryContext,
        metric_list: list[Metric],
        time_series_period: str = "weekly",
        custom_targets_query: str | None = None,
        target_list: list[HistoricalTarget] | None = None,
        replace_tables: bool = False,
    ) -> TimeSeriesResult:
        """Return a TimeSeriesResult with per-client metric values.

        Roughly equivalent to looping over :meth:`.get_single_window_data`
        with different analysis windows, and reorganising the results.

        Args:
            bq_context (BigQueryContext): BigQuery configuration and client.
            metric_list (list of mozanalysis.metric.Metric): The metrics
                to analyze.
            time_series_period (str): Period of the time series for which to
                retrieve data. Options are daily, weekly, and 28_day.
            target_list (list of mozanalysis.segments.Segment): The targets
                that define clients to be included in the analysis, based on
                inclusion in a defined user segment. For each Segment included
                in the list, client IDs are identified in the
                SegmentDataSource that based on the select_for statement in
                the Segment. Client IDs that satisfy the select_for
                statement in every Segment in the target_list will be
                returned by the query.
            custom_targets_query (str): A full SQL query to be used
                in the main query; must include a client_id column
                to join to metrics tables and an enrollment_date:

                N.B. this query's results must be uniquely keyed by
                (client_id, enrollment_date), or else your results will be
                subtly wrong.
            replace_tables (bool): If True, delete tables that exist in
                BigQuery with the same name; used to rerun analyses under
                the same experiment name with different settings

        Returns:
            A :class:`mozanalysis.experiment.TimeSeriesResult` object,
            which may be used to obtain a
            pandas DataFrame of per-client metric data, for each
            analysis window. Each DataFrame is a pandas DataFrame in
            "the standard format": one row per client, some metadata
            columns, plus one column per metric and sanity-check metric.
            Its columns (not necessarily in order):

                * enrollment date and client_id
                * [metric 1]: The client's value for the first metric in
                  ``metric_list``.
                * ...
                * [metric n]: The client's value for the nth (final)
                  metric in ``metric_list``.
        """
        last_date_full_data = add_days(
            self.start_date, self.num_dates_enrollment + self.analysis_length - 1
        )

        today = date.today().strftime("%Y-%m-%d")
        if last_date_full_data >= today:
            raise ValueError(
                f"Based on the start date, {self.start_date}"
                + f", with {self.num_dates_enrollment} days of enrollment "
                + f"and analysis of length {self.analysis_length} days, "
                + f"the last day of analysis is {last_date_full_data}"
                + ", which is in the future."
            )

        time_limits = TimeLimits.for_ts(
            self.start_date,
            last_date_full_data,
            time_series_period,
            self.num_dates_enrollment,
        )

        self._targets_sql = self.build_targets_query(
            time_limits=time_limits,
            target_list=target_list,
            custom_targets_query=custom_targets_query,
        )

        targets_table_name = sanitize_table_name_for_bq(
            "_".join(
                [
                    time_limits.last_date_data_required,
                    "targets",
                    self.experiment_name,
                    hash_ish(self._targets_sql),
                ]
            )
        )

        bq_context.run_query(self._targets_sql, targets_table_name, replace_tables)

        self._metrics_sql = self.build_metrics_query(
            metric_list=metric_list,
            time_limits=time_limits,
            targets_table=bq_context.fully_qualify_table_name(targets_table_name),
        )

        full_res_table_name = sanitize_table_name_for_bq(
            "_".join(
                [
                    time_limits.last_date_data_required,
                    self.experiment_name,
                    hash_ish(self._metrics_sql),
                ]
            )
        )

        bq_context.run_query(self._metrics_sql, full_res_table_name, replace_tables)

        return TimeSeriesResult(
            fully_qualified_table_name=bq_context.fully_qualify_table_name(
                full_res_table_name
            ),
            analysis_windows=time_limits.analysis_windows,
        )

    def build_targets_query(
        self,
        time_limits: TimeLimits,
        target_list: Segment | None = None,
        custom_targets_query: str | None = None,
    ) -> str:
        return """
        {targets_query}
        """.format(
            targets_query=custom_targets_query
            or self._build_targets_query(target_list, time_limits)
        )

    def build_metrics_query(
        self,
        metric_list: list[Metric],
        time_limits: TimeLimits,
        targets_table: str,
    ) -> str:
        """Return a SQL query for querying metric data.

        For interactive use, prefer :meth:`.get_single_window_data`
        or TODO :meth:`.get_time_series_data`, according to your use case,
        which will run the query for you and return a materialized
        dataframe.

        Args:
            metric_list (list of mozanalysis.metric.Metric):
                The metrics to analyze.
            time_limits (TimeLimits): An object describing the
                interval(s) to query
            targets_table (str): The name of the targets BigQuery table

        Returns:
            A string containing a BigQuery SQL expression.

        Building this query is the main goal of this module.
        """
        analysis_windows_query = self._build_analysis_windows_query(
            time_limits.analysis_windows
        )

        metrics_columns, metrics_joins = self._build_metrics_query_bits(
            metric_list, time_limits
        )

        return """
        WITH analysis_windows AS (
            {analysis_windows_query}
        ),
        targets AS (
            SELECT
                t.*,
                aw.*
            FROM `{targets_table}` t
            CROSS JOIN analysis_windows aw
        )
        SELECT
            targets.*,
            {metrics_columns}
        FROM targets
        {metrics_joins}
        """.format(
            analysis_windows_query=analysis_windows_query,
            metrics_columns=",\n        ".join(metrics_columns),
            metrics_joins="\n".join(metrics_joins),
            targets_table=targets_table,
        )

    @staticmethod
    def _build_analysis_windows_query(analysis_windows: tuple[AnalysisWindow]) -> str:
        """Return SQL to construct a table of analysis windows.

        To query a time series, we construct a table of analysis windows
        and cross join it with the targets table to get one row per
        pair of client and analysis window.

        This method writes the SQL to define the analysis window table.
        """
        return "\n        UNION ALL\n        ".join(
            f"""(SELECT {aw.start} AS analysis_window_start,
                {aw.end} AS analysis_window_end)"""
            for aw in analysis_windows
        )

    def _partition_by_data_source(
        self, metric_list: list[Metric]
    ) -> dict[DataSource | SegmentDataSource, list[Metric | Segment]]:
        """Return a dict mapping data sources to target/metric lists."""
        data_sources = {m.data_source for m in metric_list}

        return {
            ds: [m for m in metric_list if m.data_source == ds] for ds in data_sources
        }

    def _build_targets_query(
        self, target_list: list[Segment], time_limits: TimeLimits
    ) -> str:
        target_queries = []
        target_columns = []
        dates_columns = []
        target_joins = []

        for i, t in enumerate(target_list):
            query_for_target = t.data_source.build_query_target(
                t, time_limits, from_expr_dataset=self.app_id
            )

            target_queries.append(
                f"""
        ds_{i} AS (
                {query_for_target}
            ),"""
            )

            target_columns.append(
                f"""
                    ,ds_{i}.{t.name}
                    ,ds_{i}.target_first_date as {t.name}_first_date
                """
            )

            if i != 0:
                target_joins.append(
                    f"""
                    INNER JOIN ds_{i}
                        ON ds_{i}.client_id = ds_0.client_id
                        AND ds_{i}.target_first_date <= ds_0.target_last_date
                        AND ds_{i}.target_last_date >= ds_0.target_first_date
                        """
                )

            dates_columns.append(f"{t.name}_first_date")

        target_def = "WITH" + " ".join(q for q in target_queries)

        joined_query = """
        joined AS (
            SELECT
                ds_0.client_id
                {target_columns}
            FROM
                ds_0
                {target_joins}
            ),""".format(
            target_columns=" ".join(c for c in target_columns),
            target_joins=" ".join(j for j in target_joins),
        )

        unpivot_join = """
         unpivoted AS (
                SELECT * FROM joined
                UNPIVOT(min_dates for target_date in ({target_first_dates}))
            )
        """.format(target_first_dates=", ".join(c for c in dates_columns))

        return f"""
        {target_def}
        {joined_query}
        {unpivot_join}
        SELECT
            client_id,
            max(min_dates) as enrollment_date
        FROM unpivoted
        GROUP BY client_id
        """

    def _build_metrics_query_bits(
        self, metric_list: list[Metric], time_limits: TimeLimits
    ) -> tuple[list[str], list[str]]:
        """Return lists of SQL fragments corresponding to metrics."""
        ds_metrics = self._partition_by_data_source(metric_list)
        ds_metrics = dict(ds_metrics.items())

        metrics_columns = []
        metrics_joins = []

        for i, ds in enumerate(ds_metrics.keys()):
            query_for_metrics = ds.build_query_targets(
                ds_metrics[ds],
                time_limits,
                self.experiment_name,
                analysis_length=self.analysis_length,
                continuous_enrollment=self.continuous_enrollment,
                from_expr_dataset=self.app_id,
            )
            metrics_joins.append(
                f"""    LEFT JOIN (
        {query_for_metrics}
        ) ds_{i} USING (client_id, analysis_window_start, analysis_window_end)
                """
            )

            for m in ds_metrics[ds]:
                metrics_columns.append(f"ds_{i}.{m.name}")

        return metrics_columns, metrics_joins

    def get_targets_sql(self):
        if not self._targets_sql:
            raise ValueError(
                "Target SQL not available; call `get_single_window_data` first"
            )
        return self._targets_sql

    def get_metrics_sql(self):
        if not self._metrics_sql:
            raise ValueError(
                "Metric SQL not available; call `get_single_window_data` first"
            )
        return self._metrics_sql


@attr.s(frozen=True, slots=True)
class ContinuousEnrollmentTimeLimits:
    """Expresses time limits for different kinds of analysis windows.

    Instantiated and used by the :class:`Experiment` class; end users
    should not need to interact with it.

    Do not directly instantiate: use the constructors provided.

    There are several time constraints needed to specify a valid query
    for experiment data:

        * When did enrollments start?
        * When did enrollments stop?
        * How long after enrollment does the analysis window start?
        * How long is the analysis window?

    Even if these four quantities are specified directly, it is
    important to check that they are consistent with the available
    data - i.e. that we have data for the entire analysis window for
    every enrollment.

    Furthermore, there are some extra quantities that are useful for
    writing efficient queries:

        * What is the first date for which we need data from our data
          source?
        * What is the last date for which we need data from our data
          source?

    Instances of this class store all these quantities and do validation
    to make sure that they're consistent. The "store lots of overlapping
    state and validate" strategy was chosen over "store minimal state
    and compute on the fly" because different state is supplied in
    different contexts.
    """

    first_enrollment_date = attr.ib(type=str)
    last_enrollment_date = attr.ib(type=str)

    first_date_data_required = attr.ib(type=str)
    last_date_data_required = attr.ib(type=str)

    analysis_windows = attr.ib(type=tuple[AnalysisWindow])

    @classmethod
    def for_single_analysis_window(
        cls,
        first_date_full_data: str,
        analysis_length_dates: str,
    ) -> ContinuousEnrollmentTimeLimits:
        """Return a ``ContinuousEnrollmentTimeLimits`` instance with the following
        parameters:

        Args:
            first_enrollment_date (str): First date on which enrollment
                events were received; the start date of the experiment.
            last_date_full_data (str): The most recent date for which we
                have complete data, e.g. '2019-03-22'. If you want to ignore
                all data collected after a certain date (e.g. when the
                experiment recipe was deactivated), then do that here.
            analysis_start_days (int): the start of the analysis window,
                measured in 'days since the client enrolled'. We ignore data
                collected outside this analysis window.
            analysis_length_days (int): the length of the analysis window,
                measured in days.
            num_dates_enrollment (int, optional): Only include this many days
                of enrollments. If ``None`` then use the maximum number of days
                as determined by the metric's analysis window and
                ``last_date_full_data``. Typically ``7n+1``, e.g. ``8``. The
                factor ``7`` removes weekly seasonality, and the ``+1``
                accounts for the fact that enrollment typically starts a few
                hours before UTC midnight.
        """
        analysis_window = AnalysisWindow(0, analysis_length_dates - 1)

        last_date_data_required = add_days(first_date_full_data, analysis_length_dates)

        tl = cls(
            first_enrollment_date=first_date_full_data,
            last_enrollment_date=last_date_data_required,
            first_date_data_required=first_date_full_data,
            last_date_data_required=last_date_data_required,
            analysis_windows=(analysis_window,),
        )
        return tl



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from dataclasses import dataclass

import numpy as np
import pandas as pd
import pytest
from mozanalysis.config import ConfigLoader
from mozanalysis.frequentist_stats.sample_size import (
    empirical_effect_size_sample_size_calc,
    sample_size_curves,
    z_or_t_ind_sample_size_calc,
)

search_clients_daily = ConfigLoader.get_data_source(
    "search_clients_engines_sources_daily", "firefox_desktop"
)
uri_count = ConfigLoader.get_metric("uri_count", "firefox_desktop")


@pytest.fixture()
def fake_ts_result():
    class FakeTimeSeriesResult:
        def get_aggregated_data(self, metric_list, aggregate_function, **kwargs):
            periods = [0, 7, 14]
            values = [5, 2, 1] if aggregate_function == "AVG" else [1, 2, 1]

            df = pd.DataFrame({"analysis_window_start": periods})
            for m in metric_list:
                df[m.name] = values
            return df, 1000

    return FakeTimeSeriesResult()


def test_sample_size_calc_desktop():
    df = pd.DataFrame(
        {
            search_clients_daily.name: np.random.normal(size=100),
            uri_count.name: np.random.normal(size=100),
        }
    )

    res = z_or_t_ind_sample_size_calc(df, [search_clients_daily, uri_count])

    assert all(c in res for c in df.columns)

    assert res[search_clients_daily.name]["sample_size_per_branch"] > 1000000
    assert res[uri_count.name]["sample_size_per_branch"] > 1000000


def test_empirical_effect_size_sample_size_calc(fake_ts_result):
    @dataclass
    class FakeMetric:
        name: str

    metric_list = [FakeMetric(name="metric1"), FakeMetric(name="metric2")]

    result = empirical_effect_size_sample_size_calc(
        res=fake_ts_result, bq_context=None, metric_list=metric_list
    )

    assert len(result) == 2
    for m in ["metric1", "metric2"]:
        r = result[m]
        assert r["effect_size"]["value"] == 3
        assert r["effect_size"]["period_start_day"] == 7
        assert r["mean"]["value"] == 5
        assert r["mean"]["period_start_day"] == 0
        assert r["std_dev"]["value"] == 2
        assert r["std_dev"]["period_start_day"] == 7
        assert r["relative_effect_size"] == 0.6
        assert r["number_of_clients_targeted"] == 1000
        np.testing.assert_allclose(r["sample_size_per_branch"], 8.44, atol=0.01)
        np.testing.assert_allclose(
            r["population_percent_per_branch"], 0.844, atol=0.001
        )


def test_empirical_effect_size_results_holder(fake_ts_result):
    """checks the dataframe property doesn't throw an error and has
    the expected columns in the case where a weekly mean is not generated"""

    @dataclass
    class FakeMetric:
        name: str

    metric_list = [FakeMetric(name="metric1"), FakeMetric(name="metric2")]

    result = empirical_effect_size_sample_size_calc(
        res=fake_ts_result, bq_context=None, metric_list=metric_list
    )

    result_df = result.dataframe
    expected_columns = {
        "relative_effect_size",
        "effect_size_value",
        "mean_value",
        "std_dev_value",
        "sample_size_per_branch",
        "population_percent_per_branch",
        "effect_size_period",
        "mean_period",
        "std_dev_period",
    }
    assert set(result_df.columns) == expected_columns

    # check that style runs successfully
    _ = result.get_styled_dataframe()


def test_curve_results_holder():
    """this test ensures the results_holder object has properly formatted attributes
    and is backward compatible"""
    df = pd.DataFrame(
        {
            search_clients_daily.name: np.random.normal(size=100),
            uri_count.name: np.random.normal(size=100),
        }
    )

    metrics = [search_clients_daily, uri_count]
    effect_size = np.arange(0.01, 0.11, 0.01)
    res = sample_size_curves(
        df,
        metrics,
        solver=z_or_t_ind_sample_size_calc,
        effect_size=effect_size,
        power=0.8,
        alpha=0.05,
        outlier_percentile=99.5,
    )
    metric_names = {el.name for el in metrics}

    # set should only have 2 values: "search_clients_daily" "uri_count"
    assert len(res) == len(metric_names)
    assert set(res.keys()) == metric_names
    assert len(res.values()) == len(metric_names)
    iter_keys = []
    for key, val in res.items():
        iter_keys.append(key)
        assert set(val.columns) == {
            "effect_size",
            "sample_size_per_branch",
            "population_percent_per_branch",
            "number_of_clients_targeted",
        }
        assert set(val["effect_size"]) == set(effect_size)
    assert len(iter_keys) == len(metric_names)
    assert set(iter_keys) == metric_names


def test_curve_results_holder_dataframe():
    """this test ensures the dataframe property function works as expected"""
    df = pd.DataFrame(
        {
            search_clients_daily.name: np.random.normal(size=100),
            uri_count.name: np.random.normal(size=100),
        }
    )

    metrics = [search_clients_daily, uri_count]
    effect_size = np.arange(0.01, 0.11, 0.01)
    res = sample_size_curves(
        df,
        metrics,
        solver=z_or_t_ind_sample_size_calc,
        effect_size=effect_size,
        power=0.8,
        alpha=0.05,
        outlier_percentile=99.5,
    )

    cols_no_stats = {
        "sample_size_per_branch",
        "population_percent_per_branch",
        "number_of_clients_targeted",
    }

    no_stats = res.dataframe
    assert set(no_stats.columns) == cols_no_stats


def test_curve_results_holder_dataframe_simvar_alpha():
    """this test ensures the dataframe property function works as expected"""
    df = pd.DataFrame(
        {
            search_clients_daily.name: np.random.normal(size=100),
            uri_count.name: np.random.normal(size=100),
        }
    )

    metrics = [search_clients_daily, uri_count]
    alpha = np.arange(0.01, 0.11, 0.01)
    res = sample_size_curves(
        df,
        metrics,
        solver=z_or_t_ind_sample_size_calc,
        effect_size=0.05,
        power=0.8,
        alpha=alpha,
        outlier_percentile=99.5,
    )

    cols_no_stats = {
        "sample_size_per_branch",
        "population_percent_per_branch",
        "number_of_clients_targeted",
    }

    no_stats = res.dataframe
    assert set(no_stats.columns) == cols_no_stats


def test_get_styled_dataframe():
    "make sure highlight_listthan will not throw an error"
    df = pd.DataFrame(
        {
            search_clients_daily.name: np.random.normal(size=100),
            uri_count.name: np.random.normal(size=100),
        }
    )

    metrics = [search_clients_daily, uri_count]
    effect_size = np.arange(0.01, 0.11, 0.01)
    res = sample_size_curves(
        df,
        metrics,
        solver=z_or_t_ind_sample_size_calc,
        effect_size=effect_size,
        power=0.8,
        alpha=0.05,
        outlier_percentile=99.5,
    )
    experiment_effect_sizes = res._params["simulated_values"]
    stats_cols = {
        "mean",
        "std",
        "mean_trimmed",
        "std_trimmed",
        "trim_change_mean",
        "trim_change_std",
    }
    _ = res.get_styled_dataframe(
        input_data=df,
        append_stats=True,
        highlight_lessthan=[(10, "green"), (20, "blue")],
    )

    with pytest.raises(
        ValueError, match="append_stats is true but no raw data was provided"
    ):
        _ = res.get_styled_dataframe(append_stats=True)

    # check that subset works
    subset_experiment_effect_sizes = experiment_effect_sizes[1:-1]
    subset_cols_with_stats = set(subset_experiment_effect_sizes) | stats_cols

    no_stats = res.get_styled_dataframe(simulated_values=subset_experiment_effect_sizes)
    assert set(no_stats.data.columns) == set(subset_experiment_effect_sizes)

    with_stats = res.get_styled_dataframe(
        input_data=df,
        append_stats=False,
        simulated_values=subset_experiment_effect_sizes,
    )
    assert set(with_stats.data.columns) == set(subset_experiment_effect_sizes)

    with_stats = res.get_styled_dataframe(
        input_data=df,
        append_stats=True,
        simulated_values=subset_experiment_effect_sizes,
    )
    assert set(with_stats.data.columns) == subset_cols_with_stats


def test_results_holder():
    df = pd.DataFrame(
        {
            search_clients_daily.name: np.random.normal(size=100),
            uri_count.name: np.random.normal(size=100),
        }
    )

    metrics = [search_clients_daily, uri_count]
    res = z_or_t_ind_sample_size_calc(df, metrics)
    metric_names = {el.name for el in metrics}

    # set should only have 2 values: "search_clients_daily" "uri_count"
    assert len(res) == len(metric_names)
    assert set(res.keys()) == metric_names
    assert len(res.values()) == len(metric_names)
    iter_keys = []
    for key, _ in res.items():
        iter_keys.append(key)
    assert len(iter_keys) == len(metric_names)
    assert set(iter_keys) == metric_names

    # make sure the  get_dataframe isn't broken
    _ = res.dataframe
    res.plot_results()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import datetime
import hashlib
from functools import reduce

import numpy as np


def all_(l):
    """Return the element-wise logical AND of `Column`s.

    Think of this as a vector-friendly version of the built-in function
    `all()`.

    Args:
        l: A list of `Column`s of booleans. Or, more generally,
            an iterable of items to reduce over.

    Returns:
        A `Column` of booleans representing the logical AND. Or,
            more generally, the result of the logical AND.
    """
    return reduce(lambda x, y: x & y, l)


def any_(l):
    """Return the element-wise logical OR of `Column`s.

    Think of this as a vector-friendly version of the built-in function
    `any()`.

    Args:
        l: A list of `Column`s of booleans. Or, more generally,
            an iterable of items to reduce over.

    Returns:
        A `Column` of booleans representing the logical OR. Or,
            more generally, the result of the logical OR.
    """
    return reduce(lambda x, y: x | y, l)


def add_days(date_string, n_days):
    """Add `n_days` days to a date string like '2019-01-01'."""
    original_date = datetime.datetime.strptime(date_string, "%Y-%m-%d")
    new_date = original_date + datetime.timedelta(days=n_days)
    return datetime.datetime.strftime(new_date, "%Y-%m-%d")


def date_sub(date_string_l, date_string_r):
    """Return the number of days between two date strings like '2019-01-01'"""
    date_l = datetime.datetime.strptime(date_string_l, "%Y-%m-%d")
    date_r = datetime.datetime.strptime(date_string_r, "%Y-%m-%d")
    return (date_l - date_r).days


def filter_outliers(branch_data, threshold_quantile):
    """Return branch_data with outliers capped.

    N.B. `branch_data` is for an individual branch: if you do it for
    the entire experiment population in whole, then you may bias the
    results.

    Args:
        branch_data: Data for one branch as a 1D ndarray or similar.
        threshold_quantile (float): Sets outliers above this
            quantile equal to the value of this quantile.

    Returns:
        branch_data with values capped at or below the threshold
        quantile.
    """
    if threshold_quantile > 1 or threshold_quantile < 0.5:
        raise ValueError("'threshold_quantile' should be close to, and <= 1")

    min_threshold = np.min(branch_data, axis=0)
    max_threshold = np.quantile(branch_data, threshold_quantile, axis=0)

    return np.clip(branch_data, min_threshold, max_threshold)


def hash_ish(string, hex_chars=12):
    """Return a crude hash of a string."""
    return hashlib.sha256(string.encode("utf-8")).hexdigest()[:hex_chars]


def get_time_intervals(
    start_date: str | datetime.datetime,
    interval: int,
    max_num_dates_enrollment: int,
) -> list[datetime.datetime]:
    """Use a start date and create end dates for enrollment intervals.
    Used to generate intervals of enrollment to calculate metrics over
    variable enrollment lengths.

    Args:
        start_date: First date of enrollment for sizing job.
        interval: Number of days to increment the enrollment end date by.
        max_num_dates_enrollment: Ceiling for the length of the enrollment
            period.

    Returns:
        date_list: List of dates where variable enrollment windows will
            end.
    """

    if isinstance(start_date, str):
        start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")

    date = start_date + datetime.timedelta(interval - 1)
    date_list = [date.date()]
    last_enrollment_date = start_date + datetime.timedelta(
        days=(max_num_dates_enrollment - 1)
    )
    while date < last_enrollment_date:
        date = date + datetime.timedelta(interval)
        date_list.append(date.date())

    date_list[len(date_list) - 1] = last_enrollment_date.date()
    return date_list



from mozanalysis.config import ConfigLoader

desktop_segments = [
    ConfigLoader.get_segment("regular_users_v3", "firefox_desktop"),
    ConfigLoader.get_segment("new_or_resurrected_v3", "firefox_desktop"),
    ConfigLoader.get_segment("weekday_regular_v1", "firefox_desktop"),
    ConfigLoader.get_segment("allweek_regular_v1", "firefox_desktop"),
    ConfigLoader.get_segment("new_unique_profiles", "firefox_desktop"),
]

desktop_metrics = [
    ConfigLoader.get_metric("active_hours", "firefox_desktop"),
    ConfigLoader.get_metric("uri_count", "firefox_desktop"),
    ConfigLoader.get_metric("search_count", "firefox_desktop"),
    ConfigLoader.get_metric("tagged_search_count", "firefox_desktop"),
    ConfigLoader.get_metric("tagged_follow_on_search_count", "firefox_desktop"),
    ConfigLoader.get_metric("ad_clicks", "firefox_desktop"),
    ConfigLoader.get_metric("searches_with_ads", "firefox_desktop"),
    ConfigLoader.get_metric("organic_search_count", "firefox_desktop"),
    ConfigLoader.get_metric("unenroll", "firefox_desktop"),
    ConfigLoader.get_metric("view_about_logins", "firefox_desktop"),
    ConfigLoader.get_metric("view_about_protections", "firefox_desktop"),
    ConfigLoader.get_metric("connect_fxa", "firefox_desktop"),
    ConfigLoader.get_metric("pocket_rec_clicks", "firefox_desktop"),
    ConfigLoader.get_metric("pocket_spoc_clicks", "firefox_desktop"),
    ConfigLoader.get_metric("days_of_use", "firefox_desktop"),
    ConfigLoader.get_metric("qualified_cumulative_days_of_use", "firefox_desktop"),
    ConfigLoader.get_metric("disable_pocket_clicks", "firefox_desktop"),
    ConfigLoader.get_metric("disable_pocket_spocs_clicks", "firefox_desktop"),
]

fenix_metrics = [
    ConfigLoader.get_metric("uri_count", "fenix"),
    ConfigLoader.get_metric("user_reports_site_issue_count", "fenix"),
    ConfigLoader.get_metric("user_reload_count", "fenix"),
    ConfigLoader.get_metric("baseline_ping_count", "fenix"),
    ConfigLoader.get_metric("metric_ping_count", "fenix"),
    ConfigLoader.get_metric("first_run_date", "fenix"),
]


firefox_ios_metrics = [
    ConfigLoader.get_metric("baseline_ping_count", "firefox_ios"),
    ConfigLoader.get_metric("metric_ping_count", "firefox_ios"),
    ConfigLoader.get_metric("first_run_date", "firefox_ios"),
]

klar_android_metrics = [
    ConfigLoader.get_metric("baseline_ping_count", "klar_android"),
    ConfigLoader.get_metric("metric_ping_count", "klar_android"),
    ConfigLoader.get_metric("first_run_date", "klar_android"),
]

klar_ios_metrics = [
    ConfigLoader.get_metric("baseline_ping_count", "klar_ios"),
    ConfigLoader.get_metric("metric_ping_count", "klar_ios"),
    ConfigLoader.get_metric("first_run_date", "klar_ios"),
]
